(* XStream compiler.
   Copyright 2006.

   INRIA Institut National de Recherche en Informatique et Automatique.
   Alain Frisch. *)


(* Parse the command line *)

let scripts = ref []
and trace_input = ref false
and trace_state = ref false
and trace_rewrite = ref false
and trace_alive = ref false
and exe = ref "a.out"
and xml = ref ""
and verbose = ref false
and quiet = ref false
and debug = ref false
and gcstats = ref false
and buffer = ref true
and concat = ref false

let spec = [
  "-trace_input", Arg.Set trace_input,
               "   Display the parsing events";
  "-trace_state", Arg.Set trace_state,
               "   Display the internal state";
  "-trace_alive", Arg.Set trace_alive,
               "   Display the number of live nodes";
  "-trace_rewrite", Arg.Set trace_rewrite,
                 " Display the rewriting steps";
  "-gcstats", Arg.Set gcstats,
           "       Display GC stats";
  "-o", Arg.Set_string exe,
     "<file>       Set output file name to <file>";
  "-run", Arg.Set_string xml,
       "<file>     Run the script on the xml document <file>";      
  "-v", Arg.Set verbose,
     "             Verbose mode";
  "-quiet", Arg.Set quiet,
         "         Run transformation in quiet mode (turn off output)";
  "-debug", Arg.Set debug,
         "         Produce human-readable OCaml source code";
  "-nobuffer", Arg.Clear buffer,
            "      Treat parsing event immediatly without buffering";
  "-concat", Arg.Set concat,
          "        Concatenate adjacent text events";
]


let () =
  Arg.parse spec 
    (fun s -> scripts := s :: !scripts)
    "Usage:   xstream <options> src1.xst ... srcn.xst"

(*******************************************************)



module Fv = Set.Make(String)
module Env = Map.Make(String)

type loc = Lexing.position * Lexing.position
let _loc = Lexing.dummy_pos, Lexing.dummy_pos

let rec forlist f i j = if i = j then [] else (f i)::(forlist f (succ i) j)
let copy x n = forlist (fun _ -> x) 0 n

type typ = Node | Basic of MLast.ctyp * MLast.expr
type arity = typ list

let constrs = Hashtbl.create 1024

let declare_constr cstr arity = Hashtbl.add constrs cstr arity

exception Error of string

let error _loc s =
  raise (Stdpp.Exc_located (_loc,Error s))

let find_arity _loc cstr n =
  try
    let a = Hashtbl.find constrs cstr in
    if List.length a != n then
      error _loc ("Arity mismatch on constructor " ^ cstr);
    a
  with Not_found ->
    let a = copy Node n in
    Hashtbl.add constrs cstr a;
    a
    

let basic_str = Basic (<:ctyp< string >>, <:expr< (fun ppf s -> Format.fprintf ppf "%S" s)  >>)
let basic_int = Basic (<:ctyp< int >>, <:expr< Format.pp_print_int >>)
let basic_bool = Basic (<:ctyp< bool >>, <:expr< Format.pp_print_bool >>)
let basic_attr = Basic (<:ctyp< attr >>, <:expr< print_attrs >>)

let builtin_constrs = [
  "elt", [ basic_str; basic_attr; Node; Node ];
  "str", [ basic_str; Node ];
  "nil", [ ];
  "apply", [ Node; Node ];

  "main", [ Node ];

  "elt1", [ basic_str; basic_attr; Node ];
  "str1", [ basic_str ];
  "concat", [ Node; Node ];
]

let () = List.iter (fun (c,a) -> declare_constr c a) builtin_constrs

let prelude = ref []


let multi_apply_pat = List.fold_left (fun a p -> <:patt< $a$ $p$ >>)
let multi_apply_exp _loc = List.fold_left (fun a p -> <:expr< $a$ $p$ >>)
let cons_name s = "X" ^ s
let cons_name_pat s = multi_apply_pat <:patt< $uid:cons_name s$ >>
let cons_name_exp s = <:expr< $uid:cons_name s$ >>
let mkfun cons = "mk_" ^ cons

(* AST *)

type pat =
  | Wildcard
  | Capture of string * pat
  | Or of pat * pat
  | Cons of string * pat list
  | Pat of MLast.patt * typ Env.t

type exp =
  | ELet of loc * loc * string * exp * exp
  | ETyp of exp * typ
  | EIf of exp * exp * exp
  | EVar of loc * string
  | ECons of loc * string * exp list
  | EExp of MLast.expr * Fv.t * typ
  | EFun of loc * string * exp
  | EMatch of loc * exp * ruls

and ruls = rul list and rul = pat * guard * exp
		    and guard = MLast.expr * Fv.t

let env_union = Env.fold Env.add
let no_guard = <:expr< True >>, Fv.empty 

let estr _loc s = EExp (<:expr< $str:s$ >>, Fv.empty, basic_str)
let pstr s = Pat (<:patt< $str:s$ >>, Env.empty)
let eint _loc s = EExp (<:expr< $int:s$ >>, Fv.empty, basic_int)
let pint s = Pat (<:patt< $int:s$ >>, Env.empty)
let cap x = Capture (x,Wildcard)
let evar x = EVar (_loc,x)

let rec type_pat t = function
  | Wildcard -> Env.empty
  | Pat (_,env) -> env
  | Capture (x,p) -> Env.add x t (type_pat t p)
  | Or (p1,p2) -> type_pat t p1 (* TODO: check p2 *)
  | Cons (s,pl) ->
      List.fold_left env_union Env.empty 
	(List.map2 type_pat (find_arity _loc s (List.length pl)) pl)

let env_find _loc x env =
  try Env.find x env
  with Not_found ->
    error _loc ("Unbound identifier " ^ x)

let rec type_exp env = function
  | ELet (_,_,x,e1,e2) -> type_exp (Env.add x (type_exp env e1) env) e2
  | ETyp (e,t) -> t
  | EIf (_,e1,_) -> type_exp env e1
  | EVar (_loc,x) -> env_find _loc x env
  | ECons _ | EFun _ -> Node
  | EExp (_,_,t) -> t
  | EMatch (_,e,(p1,_,e1)::_) -> 
      type_exp (env_union env (type_pat (type_exp env e) p1)) e1
  | EMatch _ -> assert false

let rec pat_fv = function
  | Wildcard -> Fv.empty
  | Pat (_,env) -> Env.fold (fun x t fv -> Fv.add x fv) env Fv.empty
  | Capture (x,p) -> Fv.add x (pat_fv p)
  | Or (p1,p2) -> pat_fv p1 (* TODO: check p2 *)
  | Cons (cstr,pl) -> List.fold_left Fv.union Fv.empty (List.map pat_fv pl)

let rec exp_fv = function
  | ELet (_,_,x,e1,e2) -> Fv.union (exp_fv e1) (Fv.remove x (exp_fv e2))
  | ETyp (e,_) -> exp_fv e
  | EIf (e1,e2,e3) -> Fv.union (exp_fv e1) (Fv.union (exp_fv e2) (exp_fv e3))
  | EVar (_,x) -> Fv.singleton x
  | ECons (_,_,el) -> List.fold_left Fv.union Fv.empty (List.map exp_fv el)
  | EExp (_,fv,_) -> fv
  | EFun (_,x,e) -> Fv.remove x (exp_fv e)
  | EMatch (_,e,b) -> Fv.union (exp_fv e) (ruls_fv b)

and ruls_fv ruls = List.fold_left Fv.union Fv.empty (List.map rul_fv ruls)

and rul_fv (p,(_,g),e) = Fv.union g (Fv.diff (exp_fv e) (pat_fv p))

let rec pat_depth = function
  | Wildcard | Pat _ -> 0
  | Capture (_,p) -> pat_depth p
  | Or (p1,p2) -> max (pat_depth p1) (pat_depth p2)
  | Cons (cstr,pl) -> succ (List.fold_left max 0 (List.map pat_depth pl))

let rec pat_any = function
  | Wildcard -> true
  | Capture (_,p) -> pat_any p
  | Or (p1,p2) -> pat_any p1 || pat_any p2
  | Cons _ | Pat _ -> false

let fresh_name =
  let count = ref 0 in
  fun arity -> 
    incr count; 
    let c = Printf.sprintf "_F_%i" !count in
    declare_constr c arity;
    c

let rules = ref []
let max_depth = ref 2

let constr_rules = Hashtbl.create 1024
  (* All the rules for each constructor *)

let constr_depth_table = Hashtbl.create 1024
let constr_depth c =
  try Hashtbl.find constr_depth_table c
  with Not_found -> (-1)

  (* Maximum matching depth for each constructor *)

let rec reg_rule p g e = match p with
  | Cons (c, pl) -> Hashtbl.add constr_rules c (pl,g,e)
  | Or (p1,p2) -> reg_rule p1 g e; reg_rule p2 g e
  | _ -> failwith "Toplevel capture is not allowed"

let rec reg_matched d = function
  | Or (p1,p2) -> reg_matched d p1; reg_matched d p2
  | Cons (c,pl) ->
      let i = constr_depth c in
      Hashtbl.replace constr_depth_table c (max i d);
      List.iter (reg_matched (succ d)) pl
  | Capture (_,p) -> reg_matched d p
  | Wildcard | Pat _ -> ()


(* Code generation for expressions and patterns *)

let rec make_pat = function
  | Wildcard -> <:patt< _ >>
  | Capture (x,Wildcard) -> <:patt< $lid:x$ >>
  | Capture (x,p) -> <:patt< ($make_pat p$ as $lid:x$) >>
  | Or (p1,p2) -> <:patt< $make_pat p1$ | $make_pat p2$ >>
  | Cons (cstr,pl) ->
      let k = cons_name_pat cstr (List.map make_pat pl) in
      <:patt< { kind = $k$ } >>
  | Pat (p,_) -> p

let rec make_exp env = function
  | ELet (_loc,locx,x,e1,e2) -> 
      let (t1,e1) = make_exp env e1 in
      let (t2,e2) = make_exp (Env.add x t1 env) e2 in
      t2,(match t1 with
	    | Node ->
		let px = let _loc = locx in <:patt< $lid:x$ >> in
		<:expr< let $px$ : node = $e1$ in 
	                let n : node = $e2$ in do { release $lid:x$;n } >>
            | Basic _ -> 
		let px = let _loc = locx in <:patt< $lid:x$ >> in
		<:expr< let $px$ = $e1$ in $e2$ >>)
  | ETyp (e,t) -> let (_,e) = make_exp env e in (t,e)
  | EIf (e1,e2,e3) ->
      let (t1,e1) = make_exp env e1
      and (t2,e2) = make_exp env e2
      and (t3,e3) = make_exp env e3 in
      t2,<:expr< if $e1$ then $e2$ else $e3$ >>

  | EVar (_loc,x) -> 
      let t = env_find _loc x env in
      t,(match t with Node -> <:expr< protect $lid:x$ >> 
                    | Basic _ -> <:expr< $lid:x$ >>)
  | ECons (_loc,c,el) -> 
      let el = List.map snd (List.map (make_exp env) el) in
      Node,multi_apply_exp _loc <:expr< $lid:mkfun c$ >> (<:expr< _d >> :: el)
  | EExp (e,_,t) -> (t,e)
  | EFun (_loc,x,e) ->
      let fv = Fv.elements (Fv.remove x (exp_fv e)) in
      let arity = List.map (fun x -> env_find _loc x env) fv in
      let cons = fresh_name arity in
      let p = Cons (cons, List.map cap fv) in
      add_rule (Cons ("apply", [ p ; cap x ]), no_guard, e);
      make_exp env (ECons (_loc,cons, List.map evar fv))
  | EMatch (_loc,e,b) ->
      let t = type_exp env e in
      let fv = Fv.elements (ruls_fv b) in
      let arity = t :: (List.map (fun x -> env_find _loc x env) fv) in
      let cons = fresh_name arity in
      let vs = List.map cap fv in
      List.iter (fun (p,g,e) -> add_rule (Cons (cons,  p :: vs), g, e)) b;
      make_exp env (ECons (_loc,cons, e :: List.map evar fv))

and add_rule (p,g,e) = 
  max_depth := max !max_depth (pat_depth p);
  let env = type_pat Node p in
  let (t,e) = make_exp env e in
  rules := (p,g,e) :: !rules;
  reg_matched 0 p;
  reg_rule p g e

let guard ((e,_) as g) = if g == no_guard then None else Some e

let rewrule (p,g,e) =
  let e = match e with
    | <:expr< mk_elt $_$ $tag$ $attr$ $e1$ $e2$ >> ->
      <:expr< replace_elt __n $tag$ $attr$ $e1$ $e2$ >>
    | <:expr< mk_str $_$ $s$ $e1$ >> -> <:expr< replace_str __n $s$ $e1$ >>
    | <:expr< mk_nil $_$ >> -> <:expr< replace_nil __n >> 
    | e -> <:expr< replace __n $e$ >> 
  in
  <:patt< ($make_pat p$ as __n) >> , (guard g), e

let rewrules () =
  List.rev_map rewrule !rules @ [ <:patt< _ >>, None, <:expr< () >> ]

let abstraction args body =
  List.fold_right (fun arg body -> <:expr< fun $arg$ -> $body$ >>) args body
      
let ppf = Format.std_formatter 

let all_constrs () =
  let l = Hashtbl.fold (fun cstr arity l -> (cstr,arity)::l) constrs [] in
  List.sort (fun (c1,_) (c2,_) -> compare c1 c2) l

let catchall cstr = 
  let l = Hashtbl.find_all constr_rules cstr in
  List.exists (fun (pl,g,_) -> g == no_guard && List.for_all pat_any pl) l

let builtin_kinds = [ 
  "Var", []; 
  "Killed", []; 
  "Root", [ <:ctyp< node >> ]
]

let builtin_rules = [
(*  Cons ("bool", [ cap "x" ]), no_guard,
  EExp (<:expr< do { if x then fresh Xtrue else fresh Xfalse } >>,
        Fv.singleton "x", Node); *)
]

let () = List.iter add_rule builtin_rules

let typ = function Node -> <:ctyp< node >> | Basic (t,_) -> t

let kind_typedef constrs = 
  let one_cons (c,ar) = (cons_name c, List.map typ ar) in
  let cstrs = List.map one_cons constrs in
  List.map (fun (c,a) -> _loc,c,a) (builtin_kinds @ cstrs)

let mk_args arity =
  let i = ref (-1) in
  let args = List.map (fun t -> incr i; !i,t,Printf.sprintf "x%i" !i) arity in
  let args_pat = List.map (fun _,_,s -> <:patt< $lid:s$ >>) args in
  let args_exp = List.map (fun _,_,s -> <:expr< $lid:s$ >>) args in
  args, args_pat, args_exp

let tuple_pat _loc = function
  | [] -> assert false
  | [p] -> p
  | pl -> <:patt< ($list:pl$) >>

let tuple_exp _loc = function
  | [] -> assert false
  | [e] -> e
  | el -> <:expr< ($list:el$) >>

let cstr_fun = function
  | (cstr,[]) ->
      let ruls = Hashtbl.find_all constr_rules cstr in
      if ruls != [] 
      then failwith ("Constant constructor " ^ cstr ^ " may not be rewritten")
      else  
      <:patt< $lid:mkfun cstr$ >>,  
      <:expr< fun _d -> fresh $uid:cons_name cstr$ >>
  | (cstr,arity) ->
      let args,args_pat,args_exp = mk_args arity in
      let k = multi_apply_exp _loc (cons_name_exp cstr) args_exp in
      let dolink = List.fold_right
	(fun (i,t,s) e ->
	   match t with
	     | Node -> 
		 <:expr< 
		   let () = create_link n $int:string_of_int i$ $lid:s$ 
		   in $e$ >>
	     | Basic _ -> e
	) args <:expr< n >> in
      let f = <:expr< let n = fresh $k$ in $dolink$ >> in
      let ruls = List.rev (Hashtbl.find_all constr_rules cstr) in
      let call = catchall cstr in
      let cases = 
	List.map
	  (fun (pl, g, e) ->
	     let pl = List.map make_pat pl in
	     let act = List.fold_right
	       (fun (i,t,x) e ->
		  match t with
		    | Node -> <:expr< let () = release $lid:x$ in $e$ >>
		    | Basic _ -> e) args <:expr< n >> in
	     tuple_pat _loc pl, guard g,
	     <:expr< let n : node = $e$ in $act$ >>)
	  ruls @ (if call then [] else [ <:patt< _ >>, None, f ])
      in
      let f = 
	if ruls = [] then f
	else
	  let e = <:expr< match $tuple_exp _loc args_exp$ with [$list:cases$] >> in
	  <:expr< if _d == 0 then let n : node = $f$ in do { taint n; n }
                  else let _d = pred _d in $e$ >> 
      in
      let f = abstraction args_pat <:expr< ($f$ : node) >>in
      <:patt< $lid:mkfun cstr$ >>, <:expr< fun _d -> $f$>>

let kill_constr	(cstr,arity) =
  let args,args_pat,args_exp = mk_args arity in
  let p = cons_name_pat cstr args_pat in
  let e = List.fold_right 
    (fun (i,t,x) e -> 
       match t with
	 | Node -> <:expr< let () = kill_this _d n $lid:x$ in $e$ >>
	 | Basic _ -> e)
    args <:expr< () >> in
  p, None, e
    
let kill_cases all = 
  List.map kill_constr all @ 
    [ <:patt< Killed >>, None, <:expr< assert False >>;
      <:patt< Var >>, None, <:expr< () >>;
      <:patt< Root q >>, None, <:expr< kill_this _d n q >>;
    ]

let find_stuck_constr (cstr,arity) =
  let args,args_pat,args_exp = mk_args arity in
  let p = cons_name_pat cstr args_pat in
  let e = match cstr with
    | "nil" | "elt" | "str" -> <:expr< () >>
    | _ -> <:expr< raise (Stuck n) >> in
  let e = List.fold_right 
    (fun (i,t,x) e -> 
       match t with
	 | Node -> <:expr< let () = find_stuck $lid:x$ in $e$ >>
	 | Basic _ -> e)
    args e in
  p, None, e
    
let find_stuck_cases all = 
  List.map find_stuck_constr all @ 
    [ <:patt< Killed | Var >>, None, <:expr< assert False >>;
      <:patt< Root q >>, None, <:expr< find_stuck q >>;
    ]

let print_constr (cstr,arity) =
  let args,args_pat,args_exp = mk_args arity in
  let k = cons_name_pat cstr args_pat in
  let f = 
    Printf.sprintf "%s(%s)"
      cstr
      (String.concat "," (List.map (fun _ -> "%a") args))
  in
  let p = 
    List.flatten 
      (List.map 
	 (function 
	    | (_,Node,x) -> 
		[ <:expr< print_node (d-1) >> ; <:expr< $lid:x$ >> ]
	    | (_,Basic (_,f),x) ->
		[ f; <:expr< $lid:x$ >> ]) args) in
  <:patt< { kind = $k$ } >>, None,
  multi_apply_exp _loc <:expr< Format.fprintf ppf $str:f$ >> p
  
let print_cases all =
  let all = List.filter 
    (function (("elt" | "nil"),_) -> false | _ -> true) all in
  List.map print_constr all @
    [ <:patt< {kind=Xelt(s,attr,x1,x2)} >>, None, 
      <:expr< Format.fprintf ppf "%s%a[%a]%a" s print_attrs attr 
	           (print_node (d-1)) x1
	           (print_node (d-1)) x2 >>;
      
      <:patt< {kind=Root x} >>, None, 
      <:expr< Format.fprintf ppf "Root(%a)" (print_node (d-1)) x>>;
      
      <:patt< {kind=Var} >>, None, 
      <:expr< Format.fprintf ppf "Var" >>;
      
      <:patt< {kind=Killed} >>, None, 
      <:expr< Format.fprintf ppf "Killed" >>;
      
      <:patt< {kind=Xnil} >>, None,
      <:expr< Format.fprintf ppf "()" >>;
    ]

let rec or_pat = function
  | [p] -> p
  | p::l -> <:patt< $p$ | $or_pat l$ >>
  | [] -> assert false

let is_matched all =
  let all = List.filter (fun (c,_) -> constr_depth c >= 1) all in
  let all = List.map (function
			| (c,[]) -> cons_name_pat c []
			| (c,a) -> cons_name_pat c [ <:patt< _ >> ]) all in
  if all == [] then [ <:patt< _ >>, None, <:expr< False >> ]
  else [ or_pat all, None, <:expr< True >>; 
	 <:patt< _ >>, None, <:expr< False >> ]

let gencode () =
  let all = all_constrs () in
  <:str_item< 
declare

(*
module Stk = Queue;
*)
(*
module Stk = 
  struct
    type t 'a = ref (list 'a);
    value create () = ref [];
    value add x q = q.val := [x :: q.val];
    value take q =
      match q.val with
      [ [] -> raise Not_found
      | [hd :: tl] -> do { q.val := tl; hd } ]
    ;
    value is_empty q = q.val == [];
  end
;
*)

module Stk =
  struct
    type t 'a = { q: mutable array 'a; n: mutable int };

    value create () = { q = [||]; n = 0 };
    value is_empty q = q.n == 0;
    value add x q =
	let n = q.n in
	if n = Array.length q.q then do {
	  let s = 2 * n + 10 in
	  let a = Array.create s x in
	  Array.blit q.q 0 a 0 n;
(*	  Printf.eprintf "Growing %i -> %i\n" n s; *)
	  q.q := a;
	  q.n := succ n
        } else do { q.n := succ n; q.q.(n) := x };
    value take q = let n = pred q.n in do { q.n := n; q.q.(n) };
	 
  end
;

type attr = list (string * string);
value print_attrs ppf attr = Format.fprintf ppf "{...}";

$if !trace_alive then <:str_item< value nodes_alive = ref 0 >>
 else <:str_item< declare end >>$;

type kind = [ $list:kind_typedef all$ ]
and node = { uplinks: mutable uplinks;
	     kind: mutable kind;
	     ref_cnt: mutable int;
	     tainted: mutable bool }
and uplinks = [ Nolink | Link of node and int and uplinks ];

value rec print_node d ppf n = 
  if d = 0 then Format.fprintf ppf "..." 
  else match n with [ $list:print_cases all$ ];

value print_node_depth = print_node $int: string_of_int (succ !max_depth)$;

value print_node = print_node max_int;

external set_field : 'a -> int -> 'b -> unit = "%obj_set_field";
      
value must_kill n = n.ref_cnt == 0 && n.uplinks == Nolink;
  
value rec remove_link father l = fun
 [ Nolink -> assert False
 | (Link n _ rest) as l' -> 
     if n == father then set_field l 2 rest
     else remove_link father l' rest ];

value tokill = ref [];

value rec kill _d n =
  if _d = 0 then tokill.val := [ n :: tokill.val ] else
  let () = $if !trace_alive then <:expr<decr nodes_alive>> else <:expr<()>>$ in 
  let k = n.kind in 
  let _d = pred _d in
  let () = n.kind := Killed in 
  match k with [ $list:kill_cases all$ ]
and kill_this _d father child =
  if child.ref_cnt == 0 then
  match child.uplinks with
  [ Nolink -> assert False
  | Link _ _ Nolink -> do { child.uplinks := Nolink; kill _d child }
  | Link n _ rest as l ->
        if n == father then child.uplinks := rest
	else remove_link father l rest ]
  else
  match child.uplinks with
  [ Nolink -> assert False
  | Link n _ rest as l ->
        if n == father then child.uplinks := rest
	else remove_link father l rest ];

value rec kill_loop n =
  let () = kill 1000 n in
  match tokill.val with 
  [ [hd::tl] -> let () = tokill.val := tl in kill_loop hd 
  | [] -> ()];

value kill n = kill_loop n;

value check_kill n = if must_kill n then kill n else ();

value protect n = do { n.ref_cnt := succ n.ref_cnt; n };
value release n = do { n.ref_cnt := pred n.ref_cnt; check_kill n };

value tainted = Stk.create ();

value taint n =
  if not n.tainted then do {n.tainted := True; Stk.add n tainted } else ();

value rec taint_up depth = fun
 [ Nolink -> ()
 | Link n _ next -> 
   do {
     if depth > 0 then taint_up (pred depth) n.uplinks else ();  
     taint n;
     taint_up depth next
   } ];

value fetch_tainted () =
  let n = Stk.take tainted in do { n.tainted := False; n }; 

value is_matched n = match n.kind with [$list:is_matched all$];

value replace n1 n2 = 
  let () = $if !trace_rewrite then 
    <:expr< Format.fprintf Format.err_formatter
                 "%a ---> %a@."
                 print_node n1 print_node n2 >>
  else <:expr< () >>$ in
  let rec aux = fun
    [ Link parent i Nolink as l ->
        do { set_field parent.kind i n2; set_field l 2 n2.uplinks }
    | Link parent i rest -> do { set_field parent.kind i n2; aux rest }
    | Nolink -> assert False ]
  in
  let l = n1.uplinks in
  do {
   aux l; n2.uplinks := l; n1.uplinks := Nolink; kill n1; release n2;
   if is_matched n2 then 
     taint_up $int:string_of_int (!max_depth - 2)$ n2.uplinks
   else () 
  (* Need only to mark the parents of n1, not of n2 *)
  };

value fresh k =
  let () = 
    $if !trace_alive then <:expr<incr nodes_alive>> else <:expr<()>>$ in 
  {uplinks = Nolink; kind = k; ref_cnt = 1; tainted = False};

value set_open n tag attr =
  let () = 
    $if !trace_alive then <:expr<nodes_alive.val := nodes_alive.val + 2>> 
     else <:expr<()>>$ in 
  let x1 = { uplinks=Link n 2 Nolink; kind=Var; ref_cnt=0; tainted=False }
  and x2 = { uplinks=Link n 3 Nolink; kind=Var; ref_cnt=0; tainted=False } in
  let () = n.kind := Xelt(tag,attr,x1,x2) in
  let () = taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks in
  (x1,x2);

value set_close n =
  let () = n.kind := Xnil in
  taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks;

value set_text n str =
  let () = 
    $if !trace_alive then <:expr<incr nodes_alive>> else <:expr<()>>$ in 
  let x2 = { uplinks=Link n 1 Nolink; kind=Var; ref_cnt=0; tainted=False } in
  let () = n.kind := Xstr(str,x2) in
  let () = taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks in
  x2;

value set_text_close n str =
  let () = 
    $if !trace_alive then <:expr<incr nodes_alive>> else <:expr<()>>$ in 
  let x2 = { uplinks=Link n 1 Nolink; kind=Xnil; ref_cnt=0; tainted=False } in
  let () = n.kind := Xstr(str,x2) in
  taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks;  

value set_text_open n str tag attr =
  let () = 
    $if !trace_alive then <:expr<nodes_alive.val := nodes_alive.val + 2>> 
     else <:expr<()>>$ in 
  let n' = { uplinks=Link n 1 Nolink; kind=Var; ref_cnt=0; tainted=False } in
  let x1 = { uplinks=Link n' 2 Nolink; kind=Var; ref_cnt=0; tainted=False }
  and x2 = { uplinks=Link n' 3 Nolink; kind=Var; ref_cnt=0; tainted=False } in
  let () = n'.kind := Xelt(tag,attr,x1,x2) in
  let () = n.kind := Xstr(str,n') in
  let () = taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks in
  (x1,x2);

value create_link parent i child = 
do { child.uplinks := Link parent i child.uplinks; 
     child.ref_cnt := pred child.ref_cnt };

value replace_elt n tag attr x1 x2 =
  let () = 
    $if !trace_alive then <:expr<incr nodes_alive>> else <:expr<()>>$ in 
  let () = kill n in
  let () = n.kind := Xelt (tag,attr,x1,x2) in
  let () = create_link n 2 x1
  and () = create_link n 3 x2 in
  taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks;

value replace_str n s x1 =
  let () = 
    $if !trace_alive then <:expr<incr nodes_alive>> else <:expr<()>>$ in 
  let () = kill n in
  let () = n.kind := Xstr (s,x1) in
  let () = create_link n 1 x1 in
  taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks;

value replace_nil n =
  let () = 
    $if !trace_alive then <:expr<incr nodes_alive>> else <:expr<()>>$ in 
  let () = kill n in
  let () = n.kind := Xnil in
  taint_up $int:string_of_int (!max_depth - 2)$ n.uplinks;

value open_tag s attr =  do { 
  print_char '<';
  print_string s;
  List.iter (fun (a,v) -> do { 
	       print_char ' ';
	       print_string a; print_char '='; print_char '"';
               print_string v; print_char '"' }) attr;
  print_char '>'
};

value close_tag s =
  let l = String.length s in
  let s' = String.create (l+3) in
  do { String.unsafe_set s' 0 '<';
       String.unsafe_set s' 1 '/';
       String.unsafe_set s' (l+2) '>';
       String.unsafe_blit s 0 s' 2 l;
       s' };

value root x = let n = fresh (Root x) in do { create_link n 0 x; n; };

value rec $list:List.map cstr_fun all$;
value _d = 100;
value rewrite = fun [ $list:rewrules ()$ ];

value normalize () =
  while not (Stk.is_empty tainted) do { rewrite (fetch_tainted ()) };


value dump_out s = print_string s; 
value finish () = do { print_newline (); 
		       $if !gcstats then <:expr< do { 
			 Gc.print_stat stderr;
			 Printf.eprintf "allocated_bytes: %f\n" 
			   (Gc.allocated_bytes ());
		       }
			 >>
		       else <:expr< () >>$;
		       exit 0 };

value stack = ref [];   (* Input stack *)
value first = ref True;

type out = [ Node of node and out | CloseTag of string and out | TheEnd];
value roots = ref TheEnd; (* Ouput stack *)

value rec dump_stack ppf = fun
  [ CloseTag tag rest -> 
    do { Format.fprintf ppf "Close(%s)::" tag; dump_stack ppf rest }
  | Node n rest -> 
    do { Format.fprintf ppf "Node(%a)::" print_node n; dump_stack ppf rest }
  | TheEnd -> Format.fprintf ppf "[]" ];

value trace () =
  let () = 
    $if !trace_alive 
    then <:expr< Format.fprintf Format.err_formatter 
                        "nodes:%i@." nodes_alive.val>>
    else <:expr<()>>$ in 
  let () =
    $if !trace_state 
    then <:expr<Format.fprintf Format.err_formatter "state:%a@." 
	          dump_stack roots.val>> 
    else <:expr<()>>$ in 
  ();


value rec flush_out = fun
  [ CloseTag tag rest -> do { dump_out (close_tag tag); flush_out rest }
  | Node ({ kind = Root ({kind=Xelt _|Xnil|Xstr _|Xconcat _|Xelt1 _|Xstr1 _} 
			   as n)} as r) rest ->
      do { let n = protect n in release r; flush_node n rest }
  | TheEnd -> finish ()
  | r -> roots.val := r ]
and flush_node n rest = match n.kind with  
  [ Xelt tag attr sub next ->
     do { open_tag tag attr;
          let sub = protect sub and next = protect next in
	  release n;
	  flush_node sub (CloseTag tag (Node (root next) rest)) }
  | Xconcat e1 e2 ->
     do { let e1 = protect e1 and e2 = protect e2 in
	  release n;
	  flush_node e1 (Node (root e2) rest) }
  | Xelt1 tag attr sub ->
     do { open_tag tag attr;
	  let sub = protect sub in 
	  release n; 
          flush_node sub (CloseTag tag rest) }
  | Xnil -> do { release n; flush_out rest }
  | Xstr s next -> 
     do { dump_out s; 
	  let next = protect next in
	  release n; 
	  flush_node next rest }
  | Xstr1 s -> do { dump_out s; release n; flush_out rest }
  | _ -> roots.val := Node (root n) rest ];

value do_run () = do { normalize (); flush_out roots.val; trace () };

value run = $if !buffer then <:expr<
  let timer = ref 0 in
  fun () -> do { 
    incr timer;
    if timer.val = 20 then do { do_run (); timer.val := 0 }
    else ()
  } >> else <:expr< do_run >>$;

(* Concatenate adjacent text event *)

value cdata_buf = ref [];
value add_cdata s = cdata_buf.val := [ s :: cdata_buf.val ];
value clear_cdata () = cdata_buf.val := [];

value concat l =
  let len = List.fold_left (fun len s -> len + String.length s) 0 l in
  let res = String.create len in
  let _ = List.fold_left (fun pos s ->
		    let l = String.length s in
		    let pos = pos - l in
		    let () = String.unsafe_blit s 0 res pos l in
		    pos) len l in
  res;

value start_element tag attr =
 match stack.val with 
 [ [] -> assert False
 | [ x :: rest ] ->
     if first.val then
       if x.kind == Killed then ()
       else
         let () = first.val := False in
         let x1 = fresh Var and x2 = mk_nil _d in
         let () = stack.val := [x1 :: rest] in 
	 let () = replace x (mk_elt _d tag attr x1 x2) in 
	 run ()
     else
       if x.kind == Killed then stack.val := [x; x :: rest]
       else let (x1,x2) = match cdata_buf.val with
       [ [] -> set_open x tag attr
       | [s] -> do { cdata_buf.val := []; set_text_open x s tag attr }
       | l -> do { cdata_buf.val := []; set_text_open x (concat l) tag attr }
       ] in let () = stack.val := [ x1; x2 :: rest ] in run () ];

value end_element _ =
 match stack.val with
 [ [] -> assert False
 | [ x :: rest ] ->
     if x.kind == Killed then stack.val := rest
     else do { match cdata_buf.val with
     [ [] -> set_close x
     | [s] -> do { set_text_close x s; cdata_buf.val := [] }
     | l -> do { set_text_close x (concat l); cdata_buf.val := [] }
     ]; stack.val := rest; run () }
 ];

value cdata txt =
 match stack.val with
 [ [] -> assert False
 | [ x :: rest ] ->
     if x.kind == Killed then ()
     else $if !concat then <:expr< add_cdata txt >>
           else <:expr< let x2 = set_text x txt in 
	                let () = stack.val := [x2 :: rest] in run () >>$
 ];

value parse src =
  let p = Expat.parser_create None in
  do {
    Expat.set_start_element_handler p start_element;
    Expat.set_end_element_handler p end_element;
    Expat.set_character_data_handler p cdata;
    try
      if src = "-" then
	try while True do { 
	  print_string "\n                       "; flush stdout;
	  let s = read_line () in 
	  Expat.parse p s;
	  do_run ()
	} 
	with [ End_of_file -> Expat.final p ]
      else
	let ic = open_in src in
	let buflen = 1024*4 in
	let buf = String.create buflen in
	try while True do {
	  let n = input ic buf 0 buflen in
	  if n = 0 
	  then raise End_of_file 
	  else Expat.parse_sub p buf 0 n;
	}
	with [ End_of_file -> Expat.final p ]
    with
    [ Expat.Expat_error e ->
        do {
          Printf.eprintf "Line %i, char %i: XML parse error (%s)\n"
            (Expat.get_current_line_number p)
            (Expat.get_current_column_number p) (Expat.xml_error_to_string e);
          exit 2
        } ]
  }
;

exception Stuck of node;

value rec find_stuck n = match n.kind with [$list:find_stuck_cases all$];

value find_stuck n = try do {find_stuck n; assert False} with [Stuck n -> n];

value rec dump_stuck ppf = fun
  [ Node n _ -> 
      try 
	let n = find_stuck n in
	Format.fprintf ppf "%a@." print_node_depth n
      with [ Stack_overflow -> Format.fprintf ppf "something@." ]
  | _ -> assert False ];

value () =
  let x = fresh Var in
  let () = stack.val := [x] in
  let () = roots.val := Node (root (mk_main _d x)) TheEnd in
  let () = run () in
  let () = parse Sys.argv.(1) in
  do_run ();
  Format.fprintf Format.err_formatter 
     "Computation stuck on: %a@." dump_stuck roots.val;

end
>>



(* Parser *)

let (plexer,(bolpos,lnum,fname)) = Plexer.no_quotations := false; Plexer.make_lexer ()
let gram = Grammar.gcreate plexer
let xstream = Grammar.Entry.create gram "xstream"

let parse_file src = 
  Plexer.no_quotations := false; 
  let bolpos',lnum',fname' = !bolpos,!lnum,!fname in
  bolpos := 0; lnum := 1; fname := src;
  let ic = open_in src in
  let r = Grammar.Entry.parse xstream (Stream.of_channel ic) in
  close_in ic;
  bolpos := bolpos'; lnum := lnum'; fname := fname';
  r


let match_constr _loc cstr pl =
  ignore (find_arity _loc cstr (List.length pl));
  Cons (cstr, pl)

let econs _loc cstr el =
  ignore (find_arity _loc cstr (List.length el));
  ECons (_loc,cstr, el)

let after_colon s =
  let r = String.index s ':' in
  String.sub s (succ r) (String.length s - 1 - r)


EXTEND
  GLOBAL: xstream;

  xstream: 
    [ [ ruls = LIST0 phrase; EOI -> List.flatten ruls ] ];

  phrase: 
    [ [ r = rul -> [r]
      | "include"; f = STRING -> parse_file f 
      | "caml"; q = QUOTATION ->
	  let q = after_colon q in
	  let (s,_) = Grammar.Entry.parse Pcaml.implem (Stream.of_string q) in
	  prelude := !prelude @ s;
	  []
      | "declare"; (_,c) = loc_ident; "("; ar = LIST0 typ SEP ","; ")" ->
	  declare_constr c ar; []
      | ";;" -> []
      ] ];

  typ: [ [ "_" -> Node 
	 | LIDENT "string" -> basic_str 
	 | LIDENT "int" -> basic_int
	 | LIDENT "bool" -> basic_bool
	 | q = QUOTATION ->
	     let q = after_colon q in
	     let t = Grammar.Entry.parse Pcaml.ctyp (Stream.of_string q) in
	     Basic (t, <:expr< (fun ppf s -> ()) >>)
	 ]];

  rul: [ [ p = lhs; g = OPT [ "when"; q = quot -> q ]; "->"; e = rhs -> 
	     (p,(match g with Some g -> g | None -> no_guard),e) ] ];

  lhs:
    [ [ "_" -> Wildcard
      | (_,x) = loc_ident -> Capture (x,Wildcard)
      | s = STRING -> pstr s
      | i = INT -> pint i
      | p = lhs; "as"; (_,x) = loc_ident -> Capture (x,p)
      | p1 = lhs; "|"; p2 = lhs -> Or (p1,p2)
      | "("; p = lhs; ")" -> p
      | (_,cstr) = loc_ident; "("; ch = LIST0 lhs SEP ","; ")" -> 
	  match_constr _loc cstr ch

      | (_,tag) = loc_ident; "["; a = p_attr; p1 = lhs_opt; "]"; 
	  p2 = lhs_opt -> 
	  match_constr _loc "elt" [ pstr tag; a; p1; p2 ]
      | "_"; "["; a = p_attr; p1 = lhs_opt; "]"; p2 = lhs_opt -> 
	  match_constr _loc "elt" [ Wildcard; a; p1; p2 ]
      | "%"; (_,x) = loc_ident; "["; a = p_attr; p1 = lhs_opt; 
	  "]"; p2 = lhs_opt  -> 
	  match_constr _loc "elt" [ cap x; a; p1; p2 ]

      | "("; ")" -> match_constr _loc "nil" []

      | s = STRING; p = lhs ->
	  match_constr _loc "str" [ pstr s; p ]
      | "%"; (_,x) = loc_ident; p = lhs  ->
	  match_constr _loc "str" [ cap x; p ]
      | "_"; p = lhs -> match_constr _loc "str" [ Wildcard; p ]


      | q = QUOTATION ->
	  let q = after_colon q in
	  let p = Grammar.Entry.parse Pcaml.patt (Stream.of_string q) in
	  Pat (p, Env.empty)
      ] 
    ];

  lhs_opt: 
    [ [ p = lhs -> p
      | -> match_constr _loc "nil" [] ] ];

  p_attr: [ [ "@"; (_,x) = loc_ident -> cap x
	    |  -> Wildcard
	    ] ];

  e_attr: [ [ "@"; (lx,x) = loc_ident -> EVar (lx,x)
	    |  -> EExp (<:expr< [] >>, Fv.empty, basic_int)
	    ] ];

  ident: [ [ l = LIDENT -> l | l = UIDENT -> l ] ];

  rhs:
    [ [ "let"; (l,x) = loc_ident; "="; e1 = rhs; 
	"in"; e2 = rhs -> ELet (_loc,l,x,e1,e2)
      | "let"; (l,x) = loc_ident; ":"; t = typ; "="; e1 = rhs; 
	"in"; e2 = rhs -> ELet (_loc,l,x,ETyp (e1,t),e2)
      | "if"; e1 = rhs; "then"; e2 = rhs; "else"; e3 = rhs ->
	  EIf (e1,e2,e3)
      | (lx,x) = loc_ident -> EVar (lx,x)
      | s = STRING -> estr _loc s
      | i = INT -> eint _loc i
      | "-"; i = INT -> eint _loc (string_of_int (- (int_of_string i)))
      | (_,c) = loc_ident; "("; el = LIST0 rhs SEP ","; ")" -> econs _loc c el
      | "("; e = rhs; ")" -> e

      | "("; ")" -> 
	  ECons (_loc,"nil", [])

      | (ltag,tag) = loc_ident; "["; a = e_attr; 
	  e1 = rhs_opt; "]"; e2 = rhs_opt -> 
	  ECons (_loc,"elt", [ estr ltag tag; a; e1;e2 ])
      | "%"; (lx,x) = loc_ident;
	  "["; a = e_attr; e1 = rhs_opt; "]"; e2 = rhs_opt -> 
	  ECons (_loc,"elt", [ EVar (lx,x); a; e1; e2 ])

      | "%"; (lx,x) = loc_ident; e = rhs -> 
	  ECons (_loc,"str", [ EVar (lx,x); e ])
      | s = STRING; e = rhs ->
	  ECons (_loc,"str", [ estr _loc s; e ])

      | "fun"; "["; (_,x) = loc_ident; "->"; e = rhs; "]" -> EFun (_loc,x,e)
      | "match"; e = rhs; "with"; "["; ruls = LIST0 rul SEP "|"; "]" ->
	  EMatch (_loc,e,ruls)
      | (e,fv) = quot ->
	  EExp (e, fv, basic_int)
      ]
    ];

  rhs_opt: 
    [ [ e = rhs -> e
      | -> ECons (_loc, "nil", []) ] ];


  quot: [ [ q = QUOTATION; 
	    fv = OPT [ "with"; fv = LIST1 ident SEP "," -> fv ] ->
	      let q = after_colon q in
	      let e = Grammar.Entry.parse Pcaml.expr (Stream.of_string q) in
	      let fv = match fv with 
		| Some fv -> List.fold_right Fv.add fv Fv.empty
		| None -> Fv.empty in
	      e,fv ]];
  loc_ident: [ [ x = ident -> _loc,x ] ];
	  
END

let cmd s = 
  if !verbose then Format.fprintf Format.err_formatter "%s@." s;
  if (Sys.command s) != 0 then exit 2

let main() =
  let ruls = List.flatten (List.map parse_file !scripts) in
  List.iter add_rule ruls; 
  let exe = !exe in
  let ml = exe ^ ".ml" in
  let pp = 
    if !debug then Save_print_implem.f 
    else (Pcaml.output_file := Some ml; !Pcaml.print_implem) in
  pp (!prelude @ [ gencode (), _loc ]);
  if !verbose then
    Format.fprintf Format.err_formatter "Depth = %i@." (!max_depth - 2);
  if not !debug then begin
    cmd (Printf.sprintf "ocamlfind ocamlopt -package expat -linkpkg -o %s %s" exe ml);
    Sys.remove ml;
    Sys.remove (exe ^ ".cmx");
    Sys.remove (exe ^ ".cmi");
    Sys.remove (exe ^ ".o");
    let quiet = if !quiet then "> /dev/null" else "" in
    match !xml with
      | "" -> ()
      | "-" -> cmd (Printf.sprintf "./%s - %s" exe quiet);
      | s -> cmd (Printf.sprintf "time ./%s %s %s" exe !xml quiet);
  end


let rec print_exc = function
  | Stdpp.Exc_located ((pos1,pos2),exn) -> 
      Printf.eprintf "File %S, line %i, char %i:\n" 
	pos1.Lexing.pos_fname
	pos1.Lexing.pos_lnum 
	(pos1.Lexing.pos_cnum - pos1.Lexing.pos_bol);
      print_exc exn
  | Stream.Error s ->
      Printf.eprintf "Syntax error: %s\n" s
  | Error s ->
      Printf.eprintf "Error: %s\n" s
  | exn -> Printf.eprintf "%s\n" (Printexc.to_string exn)

let () =
  try main()
  with exn -> print_exc exn
